from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroRE(Register):
    """
    Registro que contém um cadastro de restrição elétrica (RE).
    """

    IDENTIFIER = "RE"
    IDENTIFIER_DIGITS = 2
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroREHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RE.
    """

    IDENTIFIER = "RE-HORIZ-PER"
    IDENTIFIER_DIGITS = 12
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, n: datetime):
        self.data[2] = n


class RegistroRELimFormPer(Register):
    """
    Registro que contém os limites de cada restrição
    RE por patamar.
    """

    IDENTIFIER = "RE-LIM-FORM-PER"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            FloatField(size=16, decimal_digits=4, format="e"),
            FloatField(size=16, decimal_digits=4, format="e"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, v: datetime):
        self.data[2] = v

    @property
    def patamar(self) -> Optional[int]:
        return self.data[3]

    @patamar.setter
    def patamar(self, v: int):
        self.data[3] = v

    @property
    def limite_inferior(self) -> Optional[float]:
        return self.data[4]

    @limite_inferior.setter
    def limite_inferior(self, v: float):
        self.data[4] = v

    @property
    def limite_superior(self) -> Optional[float]:
        return self.data[5]

    @limite_superior.setter
    def limite_superior(self, v: float):
        self.data[5] = v



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.dlpptbmaxs import DLPPtbmaxAnos


class Dlpptbmaxs(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação das restrições
    LPP de turbinamento máximo por patamar para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `dlpptbmaxs.out`.
    """

    BLOCKS = [
        DLPPtbmaxAnos,
    ]



# Imports de módulos externos
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField


from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV


class TabelaAvlEvap(TabelaCSV):
    """
    Bloco com as informações de avaliação da representação linear
    para a evaporação.
    """

    BEGIN_PATTERN = "-----;-----;"
    LINE_MODEL = Line(
        [
            IntegerField(size=5),
            IntegerField(size=5),
            LiteralField(size=14),
            FloatField(size=10, decimal_digits=2),
            FloatField(size=10, decimal_digits=2),
            FloatField(size=10, decimal_digits=2),
            FloatField(size=10, decimal_digits=3),
            FloatField(size=10, decimal_digits=2),
        ],
        delimiter=";",
    )
    COLUMN_NAMES = [
        "periodo",
        "codigo_usina",
        "nome_usina",
        "volume_armazenado_hm3",
        "evaporacao_calculada_hm3",
        "evaporacao_modelo_hm3",
        "desvio_absoluto_hm3",
        "desvio_percentual",
    ]
    END_PATTERN = ""



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicafte import RegistroEolicaFTE, RegistroPEEFTE


class EolicaFTE(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às funções de
    transferência vento-potência.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroEolicaFTE, RegistroPEEFTE]

    def eolica_funcao_producao(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        coeficiente_linear: Optional[float] = None,
        coeficiente_angular: Optional[float] = None,
    ) -> Optional[Union[RegistroEolicaFTE, List[RegistroEolicaFTE]]]:
        """
        Obtém um registro que contém a função de produção vento-geração
        para um período de tempo.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param data_inicial: a data inicial de validade para a função
        :type data_inicial: datetime | None
        :param data_final: a data final de validade para a função
        :type data_final: datetime | None
        :param coeficiente_linear: o coeficiente linear
        :type coeficiente_linear: float | None
        :param coeficiente_angular: o coeficiente angular
        :type coeficiente_angular: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaFTE` |
            list[:class:`RegistroEolicaFTE`] | None
        """
        return self.data.get_registers_of_type(
            RegistroEolicaFTE,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            coeficiente_linear=coeficiente_linear,
            coeficiente_angular=coeficiente_angular,
        )

    def pee_fpvp_lin_pu_per(
        self,
        codigo_pee: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        coeficiente_linear: Optional[float] = None,
        coeficiente_angular: Optional[float] = None,
    ) -> Optional[Union[RegistroPEEFTE, List[RegistroPEEFTE]]]:
        """
        Obtém um registro que contém a função de produção vento-geração
        para um período de tempo para um PEE, em p.u.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param data_inicial: a data inicial de validade para a função
        :type data_inicial: datetime | None
        :param data_final: a data final de validade para a função
        :type data_final: datetime | None
        :param coeficiente_linear: o coeficiente linear
        :type coeficiente_linear: float | None
        :param coeficiente_angular: o coeficiente angular
        :type coeficiente_angular: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEFTE` |
            list[:class:`RegistroPEEFTE`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPEEFTE,
            codigo_pee=codigo_pee,
            data_inicial=data_inicial,
            data_final=data_final,
            coeficiente_linear=coeficiente_linear,
            coeficiente_angular=coeficiente_angular,
        )



from inewave.nwlistcf.modelos.caso import NomeCaso
from inewave.nwlistcf import Caso

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.caso_nwlistcf_nwlistop import (
    MockNomeCaso,
    MockCaso,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_nome_caso():
    m: MagicMock = mock_open(read_data="".join(MockNomeCaso))
    b = NomeCaso()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == "arquivos.dat"


def test_atributos_encontrados_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad = Caso.read(ARQ_TESTE)
        assert ad.arquivos is not None


def test_atributos_nao_encontrados_caso():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Caso.read(ARQ_TESTE)
        assert ad.arquivos == ""


def test_eq_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad1 = Caso.read(ARQ_TESTE)
        ad2 = Caso.read(ARQ_TESTE)
        assert ad1 == ad2


def test_neq_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad1 = Caso.read(ARQ_TESTE)
        ad2 = Caso.read(ARQ_TESTE)
        ad1.arquivos = ""
        assert ad1 != ad2


def test_leitura_escrita_caso():
    m_leitura: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m_leitura):
        ad1 = Caso.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Caso.read(ARQ_TESTE)
        assert ad1 == ad2



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.forward import VariavelOperacao, SecaoDadosForward


from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Forward(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes às simulações
    forward.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosForward]
    STORAGE = "BINARY"

    def __bloco_dados(self) -> Optional[SecaoDadosForward]:
        dados = [r for r in self.data.of_type(SecaoDadosForward)]
        if len(dados) == 1:
            return dados[0]
        else:
            return None

    def __dados_variavel(self, variavel: VariavelOperacao) -> pd.DataFrame:
        dados = self.__bloco_dados()
        if dados is not None:
            return (
                dados.data.get(variavel)
                if dados is not None
                else pd.DataFrame()
            )
        else:
            return pd.DataFrame()

    @property
    def mercado_liquido(self) -> pd.DataFrame:
        """
        A tabela com os valores do mercado líquido de energia
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - submercado (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.MERCADO)

    @property
    def energia_armazenada_absoluta_inicial(self) -> pd.DataFrame:
        """
        A tabela com os valores do EARM inicial do estágio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL
        )

    @property
    def energia_natural_afluente(self) -> pd.DataFrame:
        """
        A tabela com os valores de ENA
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL
        )

    @property
    def geracao_hidraulica_controlavel(self) -> pd.DataFrame:
        """
        A tabela com os valores de GH controlável
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL
        )

    @property
    def energia_vertida(self) -> pd.DataFrame:
        """
        A tabela com os valores de energia vertida
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.ENERGIA_VERTIDA)

    @property
    def energia_armazenada_absoluta_final(self) -> pd.DataFrame:
        """
        A tabela com os valores do EARM final do estágio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL
        )

    @property
    def energia_natural_afluente_fio_bruta(self) -> pd.DataFrame:
        """
        A tabela com os valores de ENA fio d'água bruta
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA
        )

    @property
    def energia_evaporada(self) -> pd.DataFrame:
        """
        A tabela com os valores de energia evaporada
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.ENERGIA_EVAPORADA)

    @property
    def energia_enchimento_volume_morto(self) -> pd.DataFrame:
        """
        A tabela com os valores de energia de enchimento de volume morto
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO
        )

    @property
    def geracao_termica(self) -> pd.DataFrame:
        """
        A tabela com os valores de geração térmica
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.GERACAO_TERMICA)

    @property
    def deficit(self) -> pd.DataFrame:
        """
        A tabela com os valores de déficit
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - submercado (`str`)
        - patamarDeficit (`int`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.DEFICIT)

    @property
    def valor_agua(self) -> pd.DataFrame:
        """
        A tabela com os valores da água (multiplicadores duais
        da restrição de balanço hídrico) para cada série, em R$/MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VALOR_AGUA)

    @property
    def custo_marginal_operacao(self) -> pd.DataFrame:
        """
        A tabela com os custos marginais de operação (multiplicadores duais
        da restrição de balanço de demanda) para cada série, em R$/MWh.

        - estagio (`int`)
        - cenario (`int`)
        - submercado (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.CUSTO_MARGINAL_OPERACAO)

    @property
    def geracao_hidraulica_fio_liquida(self) -> pd.DataFrame:
        """
        A tabela com os valores de geração hidráulica fio líquida
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA
        )

    @property
    def perdas_geracao_hidraulica_fio(self) -> pd.DataFrame:
        """
        A tabela com os valores de perdas na geração hidráulica fio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO
        )

    @property
    def intercambio(self) -> pd.DataFrame:
        """
        A tabela com os valores de intercâmbio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - submercadoDe (`str`)
        - submercadoPara (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.INTERCAMBIO)

    @property
    def excesso(self) -> pd.DataFrame:
        """
        A tabela com os valores de excesso
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - submercado (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.EXCESSO)

    @property
    def energia_natural_afluente_bruta(self) -> pd.DataFrame:
        """
        A tabela com os valores de ENA bruta
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA
        )

    @property
    def energia_natural_afluente_controlavel_corrigida(self) -> pd.DataFrame:
        """
        A tabela com os valores de ENA controlável corrigida
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA
        )

    @property
    def geracao_hidraulica_maxima(self) -> pd.DataFrame:
        """
        A tabela com os valores de geração hidráulica máxima
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA
        )

    @property
    def energia_afluente_controlavel_desvio(self) -> pd.DataFrame:
        """
        A tabela com os valores de energia afluente controlável
        de desvio para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO
        )

    @property
    def energia_afluente_fio_desvio(self) -> pd.DataFrame:
        """
        A tabela com os valores de energia afluente fio de desvio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO
        )

    @property
    def beneficio_intercambio(self) -> pd.DataFrame:
        """
        A tabela com os valores de benefício de intercâmbio
        para cada série, em R$/MWh.

        - estagio (`int`)
        - cenario (`int`)
        - submercadoDe (`str`)
        - submercadoPara (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.BENEFICIO_INTERCAMBIO)

    @property
    def fator_correcao_energia_natural_afluente_controlavel(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores dos fatores de correção para a
        energia natural afluente controlável
        para cada série, adimensionais.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL
        )

    @property
    def violacao_curva_aversao(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação da curva de aversão
        de volume mínimo operativo para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VIOLACAO_CURVA_AVERSAO)

    @property
    def acionamento_curva_aversao(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com o acionamento da curva de aversão
        de volume mínimo operativo para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO
        )

    @property
    def penalidade_curva_aversao(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com a penalidade por violação da curva de aversão
        de volume mínimo operativo para cada série, em R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.PENALIDADE_CURVA_AVERSAO)

    @property
    def custo_operacao(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com o custo total de operação para cada série, em 10^3 * R$.

        - estagio (`int`)
        - cenario (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.CUSTO_OPERACAO)

    @property
    def custo_geracao_termica(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com o custo de geração térmica para cada série, em 10^3 * R$.

        - estagio (`int`)
        - cenario (`int`)
        - submercado (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.CUSTO_GERACAO_TERMICA)

    @property
    def beneficio_agrupamento_intercambio(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com o benefício do agrupamento dos intercâmbios
        para cada série, em R$/MWh.

        - estagio (`int`)
        - cenario (`int`)
        - agrupamentoIntercambio (`int`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO
        )

    @property
    def energia_natural_afluente_fio(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de ENA fio
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO
        )

    @property
    def beneficio_despacho_gnl(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com o benefício do despacho térmico GNL
        para cada série, em R$/MWh.

        - estagio (`int`)
        - cenario (`int`)
        - lag (`int`)
        - patamar (`int`)
        - submercado (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.BENEFICIO_DESPACHO_GNL)

    @property
    def violacao_geracao_hidraulica_minima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação da geração hidráulica mínima
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA
        )

    @property
    def violacao_energia_vazao_minima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores da energia de vazão mínima
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA
        )

    @property
    def geracao_hidraulica_maxima_considerando_restricoes_eletricas(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores da geração hidráulica máxima considerando
        restrições elétricas para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE
        )

    @property
    def volume_armazenado_absoluto_final(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume armazenado por usina
        para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL
        )

    @property
    def geracao_hidraulica_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de geração hidráulica por usina
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.GERACAO_HIDRAULICA_USINA)

    @property
    def volume_turbinado(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume turbinado por usina
        para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VOLUME_TURBINADO)

    @property
    def volume_vertido(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume vertido por usina
        para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VOLUME_VERTIDO)

    @property
    def violacao_geracao_hidraulica_minima_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de geração hidráulica mínima
        por usina para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA
        )

    @property
    def enchimento_volume_morto_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de enchimento de volume morto
        por usina para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA
        )

    @property
    def violacao_defluencia_minima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições de
        defluência mínima por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA
        )

    @property
    def volume_desvio_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume de desvio
        por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VOLUME_DESVIO_USINA)

    @property
    def volume_desvio_positivo_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume de desvio positivo
        por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA
        )

    @property
    def volume_desvio_negativo_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume de desvio negativo
        por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA
        )

    @property
    def violacao_fpha(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições da
        FPHA por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VIOLACAO_FPHA)

    @property
    def vazao_afluente(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de vazão afluente
        por usina para cada série, em m3/s.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VAZAO_AFLUENTE)

    @property
    def vazao_incremental(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de vazão incremental
        por usina para cada série, em m3/s.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VAZAO_INCREMENTAL)

    @property
    def volume_armazenado_percentual_final(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume armazenado final
        por usina para cada série, em % do volume máximo.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL
        )

    @property
    def custo_violacao_energia_vazao_minima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de custo de violação das restrições
        de energia da vazão mínima
        para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA
        )

    @property
    def custo_energia_afluente_controlavel_desvio(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de custo da violação da meta de
        desvio da energia afluente controlável
        para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO
        )

    @property
    def custo_energia_afluente_fio_desvio(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de custo da violação da meta de
        desvio da energia afluente fio
        para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO
        )

    @property
    def custo_violacao_geracao_hidraulica_minima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de custo da violação da geração
        hidráulica mínima
        para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA
        )

    @property
    def geracao_eolica(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de geração eólica
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - pee (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.GERACAO_EOLICA)

    @property
    def velocidade_vento(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de velocidade do vento
        para cada série, em m/s.

        - estagio (`int`)
        - cenario (`int`)
        - pee (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VELOCIDADE_VENTO)

    @property
    def violacao_funcao_producao_eolica(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação da função de produção
        eólica para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - pee (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA
        )

    @property
    def violacao_defluencia_maxima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições de
        defluência máxima para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA
        )

    @property
    def violacao_turbinamento_maximo(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições de
        turbinamento máximo para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO
        )

    @property
    def violacao_turbinamento_minimo(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições de
        turbinamento mínimo para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO
        )

    @property
    def violacao_lpp_turbinamento_maximo(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições LPP de
        turbinamento máximo para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO
        )

    @property
    def violacao_lpp_defluencia_maxima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições LPP de
        defluência máxima para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA
        )

    @property
    def violacao_lpp_turbinamento_maximo_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições LPP de
        turbinamento máximo por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA
        )

    @property
    def violacao_lpp_defluencia_maxima_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de restrições LPP de
        defluência máxima por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA
        )

    @property
    def rhs_lpp_turbinamento_maximo(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores do RHS de restrições LPP de
        turbinamento máximo para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO
        )

    @property
    def rhs_lpp_defluencia_maxima(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores do RHS de restrições LPP de
        defluência máxima para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - ree (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA
        )

    @property
    def rhs_lpp_turbinamento_maximo_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores do RHS de restrições LPP de
        turbinamento máximo por usina para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA
        )

    @property
    def rhs_lpp_defluencia_maxima_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores do RHS de restrições LPP de
        defluência máxima por usina para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA
        )

    @property
    def violacao_restricoes_eletricas_especiais(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de violação de
        restrições elétricas especiais para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - restricao (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS
        )

    @property
    def custo_restricoes_eletricas_especiais(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de custo de violação de
        restrições elétricas especiais para cada série, em 10^3 R$.

        - estagio (`int`)
        - cenario (`int`)
        - restricao (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS
        )

    @property
    def volume_armazenado_absoluto_inicial(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores de volume armazenado inicial
        por usina para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL
        )

    @property
    def valor_agua_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os valores da água (multiplicadores duais
        da restrição de balanço hídrico) por usina para cada série,
        em R$/hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VALOR_AGUA_USINA)

    @property
    def volume_evaporado(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os volumes evaporados por usina para cada série,
        em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VOLUME_EVAPORADO)

    @property
    def volume_bombeado(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os volumes bombeados por estação de bombeamento
        para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - estacao (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(VariavelOperacao.VOLUME_BOMBEADO)

    @property
    def consumo_energia_estacao_bombeamento(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os consumos de energia por estação de bombeamento
        para cada série, em MWmes.

        - estagio (`int`)
        - cenario (`int`)
        - estacao (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO
        )

    @property
    def volume_canal_desvio_usina(
        self,
    ) -> pd.DataFrame:
        """
        A tabela com os volumes de devio por canal de desvio
        para cada série, em hm3.

        - estagio (`int`)
        - cenario (`int`)
        - usina (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame
        """
        return self.__dados_variavel(
            VariavelOperacao.VOLUME_CANAL_DESVIO_USINA
        )



from inewave.config import (
    MAX_ANOS_ESTUDO,
    MAX_PATAMARES,
    MAX_SUBMERCADOS,
    MESES_DF,
)

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore
from inewave._utils.formatacao import (
    prepara_vetor_anos_tabela,
    prepara_valor_ano,
    repete_vetor,
)


class BlocoNumeroPatamares(Section):
    """
    Bloco com o número de patamares de carga considerados
    no caso.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(2, 1),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoNumeroPatamares):
            return False
        bloco: BlocoNumeroPatamares = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())[0]

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, int):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )
        file.write(self.__linha.write([self.data]))


class BlocoDuracaoPatamar(Section):
    """
    Bloco com a duração de cada patamar por mês
    de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "SUBSISTEMA"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campos_duracao: List[Field] = [
            FloatField(6, 6 + i * 8, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_duracao)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoDuracaoPatamar):
            return False
        bloco: BlocoDuracaoPatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "data": prepara_vetor_anos_tabela(anos),
                    "patamar": repete_vetor(patamares),
                    "valor": tabela.flatten(),
                }
            )
            df = df.fillna(1.0)
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        ano_atual = 0
        patamar_atual = 0
        patamares: List[int] = []
        anos: List[int] = []
        tabela = np.zeros((MAX_SUBMERCADOS * MAX_ANOS_ESTUDO, len(MESES_DF)))
        while True:
            ultima_linha = file.tell()
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoDuracaoPatamar.FIM_BLOCO in linha:
                file.seek(ultima_linha)
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                    patamar_atual = 1
                anos.append(ano_atual)
                patamares.append(patamar_atual)
                tabela[i, :] = dados[1:]
                patamar_atual += 1
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        df = self.data.copy()
        df["ano"] = df.apply(lambda linha: linha["data"].year, axis=1)
        for ano in df["ano"].unique():
            for patamar in sorted(df["patamar"].unique()):
                df_ano_patamar = df.loc[
                    (df["ano"] == ano) & (df["patamar"] == patamar)
                ]
                df_ano_patamar = df_ano_patamar.sort_values(["data"])
                ano_linha = prepara_valor_ano(ano) if patamar == 1 else None
                valores = df_ano_patamar["valor"].tolist()
                file.write(self.__linha.write([ano_linha] + valores))


class BlocoCargaPatamar(Section):
    """
    Bloco com a carga de cada patamar por mês
    de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(4, 0)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campo_carga: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campo_carga)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCargaPatamar):
            return False
        bloco: BlocoCargaPatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "codigo_submercado": repete_vetor(submercados),
                    "data": prepara_vetor_anos_tabela(anos),
                    "patamar": repete_vetor(patamares),
                    "valor": tabela.flatten(),
                }
            )
            df = df.fillna(1.0)
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        ano_atual = 0
        patamar_atual = 0
        submercados: List[int] = []
        anos: List[int] = []
        patamares: List[int] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_ANOS_ESTUDO * MAX_PATAMARES,
                len(MESES_DF),
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoCargaPatamar.FIM_BLOCO in linha[:4]:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 8:
                subsis_atual = self.__linha_subsis.read(linha)[0]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                    patamar_atual = 1
                submercados.append(subsis_atual)
                anos.append(ano_atual)
                patamares.append(patamar_atual)
                tabela[i, :] = dados[1:]
                patamar_atual += 1
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema = 0

        df = self.data.copy()
        df["ano"] = df.apply(lambda linha: linha["data"].year, axis=1)
        for _, linha_submercado in (
            df[["codigo_submercado", "ano"]].drop_duplicates().iterrows()
        ):
            for patamar in sorted(df["patamar"].unique()):
                df_merc = df.loc[
                    (
                        df["codigo_submercado"]
                        == linha_submercado["codigo_submercado"]
                    )
                    & (df["ano"] == linha_submercado["ano"])
                    & (df["patamar"] == patamar)
                ]
                df_merc = df_merc.sort_values(["data"])
                if linha_submercado["codigo_submercado"] != ultimo_subsistema:
                    ultimo_subsistema = linha_submercado["codigo_submercado"]
                    file.write(
                        self.__linha_subsis.write(
                            linha_submercado[
                                [
                                    "codigo_submercado",
                                ]
                            ].tolist()
                        )
                    )
                ano_linha = (
                    prepara_valor_ano(linha_submercado["ano"])
                    if patamar == 1
                    else None
                )
                valores = df_merc["valor"].tolist()
                file.write(self.__linha.write([ano_linha] + valores))
        file.write(BlocoCargaPatamar.FIM_BLOCO + "\n")


class BlocoIntercambioPatamarSubsistemas(Section):
    """
    Bloco com o fator de correção do intercâmbio entre subsistemas para
    cada patamar, por mês de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1), IntegerField(3, 5)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campos_intercambios: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_intercambios)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoIntercambioPatamarSubsistemas):
            return False
        bloco: BlocoIntercambioPatamarSubsistemas = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "submercado_de": repete_vetor(submercados_de),
                    "submercado_para": repete_vetor(submercados_para),
                    "data": prepara_vetor_anos_tabela(anos),
                    "patamar": repete_vetor(patamares),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_de_atual = 0
        subsis_para_atual = 0
        ano_atual = 0
        patamar_atual = 0
        submercados_de: List[int] = []
        submercados_para: List[int] = []
        patamares: List[int] = []
        anos: List[int] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF),
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                len(linha) < 3
                or BlocoIntercambioPatamarSubsistemas.FIM_BLOCO in linha[:4]
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 12:
                dados = self.__linha_subsis.read(linha)
                subsis_de_atual = dados[0]
                subsis_para_atual = dados[1]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                    patamar_atual = 1
                submercados_de.append(subsis_de_atual)
                submercados_para.append(subsis_para_atual)
                anos.append(ano_atual)
                patamares.append(patamar_atual)
                tabela[i, :] = dados[1:]
                patamar_atual += 1
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema_de = 0
        ultimo_subsistema_para = 0

        df = self.data.copy()
        df["ano"] = df.apply(lambda linha: linha["data"].year, axis=1)
        for _, linha_submercado in (
            df[["submercado_de", "submercado_para", "ano"]]
            .drop_duplicates()
            .iterrows()
        ):
            for patamar in sorted(df["patamar"].unique()):
                df_merc = df.loc[
                    (df["submercado_de"] == linha_submercado["submercado_de"])
                    & (
                        df["submercado_para"]
                        == linha_submercado["submercado_para"]
                    )
                    & (df["ano"] == linha_submercado["ano"])
                    & (df["patamar"] == patamar)
                ]
                df_merc = df_merc.sort_values(["data"])
                if any(
                    [
                        linha_submercado["submercado_de"]
                        != ultimo_subsistema_de,
                        linha_submercado["submercado_para"]
                        != ultimo_subsistema_para,
                    ]
                ):
                    ultimo_subsistema_de = linha_submercado["submercado_de"]
                    ultimo_subsistema_para = linha_submercado[
                        "submercado_para"
                    ]
                    file.write(
                        self.__linha_subsis.write(
                            linha_submercado[
                                [
                                    "submercado_de",
                                    "submercado_para",
                                ]
                            ].tolist()
                        )
                    )
                ano_linha = (
                    prepara_valor_ano(linha_submercado["ano"])
                    if patamar == 1
                    else None
                )
                valores = df_merc["valor"].tolist()
                file.write(self.__linha.write([ano_linha] + valores))
        file.write(BlocoIntercambioPatamarSubsistemas.FIM_BLOCO + "\n")


class BlocoUsinasNaoSimuladas(Section):
    """
    Bloco com a geração de usinas não simuladas em P.U. para
    cada patamar, por mês de estudo, extraído do arquivo `patamar.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1), IntegerField(3, 5)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campos_usinas: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_usinas)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUsinasNaoSimuladas):
            return False
        bloco: BlocoUsinasNaoSimuladas = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "codigo_submercado": repete_vetor(submercados),
                    "indice_bloco": repete_vetor(blocos),
                    "data": prepara_vetor_anos_tabela(anos),
                    "patamar": repete_vetor(patamares),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        patamar_atual = 0
        bloco_atual = 0
        ano_atual = 0
        submercados: List[int] = []
        blocos: List[int] = []
        patamares: List[int] = []
        anos: List[int] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF),
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 12:
                dados = self.__linha_subsis.read(linha)
                subsis_atual = dados[0]
                bloco_atual = dados[1]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                    patamar_atual = 1
                submercados.append(subsis_atual)
                patamares.append(patamar_atual)
                blocos.append(bloco_atual)
                anos.append(ano_atual)
                tabela[i, :] = dados[1:]
                patamar_atual += 1
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema = 0
        ultimo_bloco = 0

        df = self.data.copy()
        df["ano"] = df.apply(lambda linha: linha["data"].year, axis=1)
        for _, linha_submercado in (
            df[["codigo_submercado", "indice_bloco", "ano"]]
            .drop_duplicates()
            .iterrows()
        ):
            for patamar in sorted(df["patamar"].unique()):
                df_merc = df.loc[
                    (
                        df["codigo_submercado"]
                        == linha_submercado["codigo_submercado"]
                    )
                    & (df["indice_bloco"] == linha_submercado["indice_bloco"])
                    & (df["ano"] == linha_submercado["ano"])
                    & (df["patamar"] == patamar)
                ]
                df_merc = df_merc.sort_values(["data"])
                if any(
                    [
                        linha_submercado["codigo_submercado"]
                        != ultimo_subsistema,
                        linha_submercado["indice_bloco"] != ultimo_bloco,
                    ]
                ):
                    ultimo_subsistema = linha_submercado["codigo_submercado"]
                    ultimo_bloco = linha_submercado["indice_bloco"]
                    file.write(
                        self.__linha_subsis.write(
                            linha_submercado[
                                [
                                    "codigo_submercado",
                                    "indice_bloco",
                                ]
                            ].tolist()
                        )
                    )
                ano_linha = (
                    prepara_valor_ano(linha_submercado["ano"])
                    if patamar == 1
                    else None
                )
                valores = df_merc["valor"].tolist()
                file.write(self.__linha.write([ano_linha] + valores))
        file.write(BlocoIntercambioPatamarSubsistemas.FIM_BLOCO + "\n")



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.restricaovazao import (
    RegistroRHQ,
    RegistroRHQHorizPer,
    RegistroRHQLsLPPVoli,
)


class RestricaoVazao(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro
    das restrições lineares por partes no domínio de vazão (RHQ).
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroRHQLsLPPVoli,
        RegistroRHQHorizPer,
        RegistroRHQ,
    ]

    def rhq(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
    ) -> Optional[Union[RegistroRHQ, List[RegistroRHQ]]]:
        """
        Obtém um registro que cadastra uma usina restrição linear por
        partes de vazão (UHE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQ` |
            list[:class:`RegistroRHQ`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHQ,
            codigo_restricao=codigo_restricao,
            formula=formula,
        )

    def rhq_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[Union[RegistroRHQHorizPer, List[RegistroRHQHorizPer]]]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição linear por partes de vazão.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicial: data inicial de validade da restrição
        :type data_inicial: datetime | None
        :param data_final: data final de validade da restrição
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQHorizPer` |
            list[:class:`RegistroRHQHorizPer`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHQHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def rhq_ls_lpp_voli(
        self,
        codigo_restricao: Optional[int] = None,
        indice_reta: Optional[int] = None,
        coeficiente_angular: Optional[float] = None,
        coeficiente_linear: Optional[float] = None,
    ) -> Optional[Union[RegistroRHQLsLPPVoli, List[RegistroRHQLsLPPVoli]]]:
        """
        Obtém um registro que cadastra as retas da restrição linear
        por partes de vazão.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param indice_reta: índice da reta definida
        :type indice_reta: int | None
        :param coeficiente_angular: coeficiente angular da reta
        :type coeficiente_angular: float | None
        :param coeficiente_linear: coeficiente linear da reta
        :type coeficiente_linear: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQLsLPPVoli` |
            list[:class:`RegistroRHQLsLPPVoli`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRHQLsLPPVoli,
            codigo_restricao=codigo_restricao,
            indice_reta=indice_reta,
            coeficiente_angular=coeficiente_angular,
            coeficiente_linear=coeficiente_linear,
        )



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)

from inewave.nwlistop.modelos.edesvcm import EdesvcmAnos


class Edesvcm(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes às energias
    de desvio de água controlável por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `edesvcm00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        EdesvcmAnos,
    ]
